# ToUpload/generators/generator_utils.py
# -*- coding: utf-8 -*-
import re
import os
import json

# --- Importar format_variable_name desde processors ---
try:
    from processors.processor_utils import format_variable_name
except ImportError:
    print("Advertencia: No se pudo importar 'format_variable_name' desde processors.processor_utils.")
    print("Usando una implementación local básica.")
    def format_variable_name(name): # Fallback
        if not name: return "_INVALID_NAME_"
        if name.startswith('"') and name.endswith('"'): return name
        prefix = "#" if name.startswith("#") else ""
        if prefix: name = name[1:]
        if name and name[0].isdigit(): name = "_" + name
        name = re.sub(r"[^a-zA-Z0-9_]", "_", name)
        return prefix + name
# --- Fin Fallback ---

# --- format_scl_start_value (Sin cambios respecto a la versión anterior) ---
def format_scl_start_value(value, datatype):
    if value is None: return None
    datatype_lower = datatype.lower() if isinstance(datatype, str) else ""
    value_str = str(value)
    is_complex_type = ('"' in datatype_lower or 'array' in datatype_lower or 'struct' in datatype_lower or datatype_lower not in ["bool", "int", "dint", "sint", "usint", "uint", "udint", "lint", "ulint", "byte", "word", "dword", "lword", "real", "lreal", "time", "ltime", "s5time", "date", "dt", "dtl", "tod", "string", "char", "wstring", "wchar", "variant"])
    if is_complex_type:
        if isinstance(value, dict): return "/* Array init TBD */"
        if re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', value_str): return value_str
        if value_str == '0': return '0'
        if value_str.lower() == 'false': return 'FALSE'
        if value_str == "''" or value_str == "": return "''"
        # print(f"INFO: Start value '{value_str}' for complex type '{datatype}' skipped.")
        return None
    value_str_unquoted = value_str;
    if value_str.startswith('"') and value_str.endswith('"') and len(value_str) > 1: value_str_unquoted = value_str[1:-1]
    elif value_str.startswith("'") and value_str.endswith("'") and len(value_str) > 1: value_str_unquoted = value_str[1:-1]
    if any(t in datatype_lower for t in ["int","byte","word","dint","dword","lint","lword","sint","usint","uint","udint","ulint"]):
        try: return str(int(value_str_unquoted))
        except ValueError: return value_str_unquoted if re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', value_str_unquoted) else None
    elif "bool" in datatype_lower: val_low = value_str_unquoted.lower(); return "TRUE" if val_low == 'true' or val_low == '1' else ("FALSE" if val_low == 'false' or val_low == '0' else (value_str_unquoted if re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', value_str_unquoted) else "FALSE"))
    elif "string" in datatype_lower or "char" in datatype_lower: escaped_value = value_str_unquoted.replace("'", "''"); prefix = "WSTRING#" if "wstring" in datatype_lower else ("WCHAR#" if "wchar" in datatype_lower else ""); return f"{prefix}'{escaped_value}'"
    elif "real" in datatype_lower or "lreal" in datatype_lower:
        try: f_val = float(value_str_unquoted); s_val = "{:.7g}".format(f_val); return s_val + (".0" if "." not in s_val and "e" not in s_val.lower() else "")
        except ValueError: return value_str_unquoted if re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', value_str_unquoted) else None
    elif "time" in datatype_lower:
        prefix, val_to_use = "", value_str_unquoted
        if val_to_use.upper().startswith("T#"): prefix, val_to_use = "T#", val_to_use[2:]
        elif val_to_use.upper().startswith("LT#"): prefix, val_to_use = "LT#", val_to_use[3:]
        elif val_to_use.upper().startswith("S5T#"): prefix, val_to_use = "S5T#", val_to_use[4:]
        if re.match(r'^-?(\d+d_)?(\d+h_)?(\d+m_)?(\d+s_)?(\d+ms)?$', val_to_use, re.IGNORECASE): target_prefix = "S5T#" if "s5time" in datatype_lower else ("LT#" if "ltime" in datatype_lower else "T#"); return f"{target_prefix}{val_to_use}"
        elif re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', value_str_unquoted): return value_str_unquoted
        else: return None
    elif any(t in datatype_lower for t in ["date", "dtl", "dt", "tod", "time_of_day"]):
         val_to_use = value_str_unquoted; prefix = ""
         if val_to_use.upper().startswith("DTL#"): prefix, val_to_use = "DTL#", val_to_use[4:]
         elif val_to_use.upper().startswith("D#"): prefix, val_to_use = "D#", val_to_use[2:]
         elif val_to_use.upper().startswith("DT#"): prefix, val_to_use = "DT#", val_to_use[3:]
         elif val_to_use.upper().startswith("TOD#"): prefix, val_to_use = "TOD#", val_to_use[4:]
         target_prefix="DTL#" if "dtl" in datatype_lower or "date_and_time" in datatype_lower else ("DT#" if "dt" in datatype_lower else ("TOD#" if "tod" in datatype_lower or "time_of_day" in datatype_lower else "D#"))
         if re.match(r'^\d{4}-\d{2}-\d{2}(-\d{2}:\d{2}:\d{2}(\.\d+)?)?$', val_to_use) or re.match(r'^\d{2}:\d{2}:\d{2}(\.\d+)?$', val_to_use): return f"{target_prefix}{val_to_use}"
         elif re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', value_str_unquoted): return value_str_unquoted
         else: return None
    else: return value_str if re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', value_str) else None


# --- generate_scl_declarations (MODIFICADO para filtrar búsqueda de UDT) ---
def generate_scl_declarations(variables, indent_level=1, project_root_dir=None):
    scl_lines = []
    indent = "  " * indent_level
    # Lista de tipos básicos simples (en minúsculas)
    basic_types = {"bool", "int", "dint", "sint", "usint", "uint", "udint", "lint", "ulint",
                   "byte", "word", "dword", "lword", "real", "lreal", "time", "ltime",
                   "s5time", "date", "dt", "dtl", "tod", "char", "wchar", "variant", "timer", "counter"} # Añadidos Timer/Counter

    # Patrones para tipos básicos parametrizados (ignorando mayúsculas/minúsculas)
    string_pattern = re.compile(r"^(W?STRING)(\[\s*\d+\s*\])?$", re.IGNORECASE)
    array_pattern = re.compile(r'^(Array\[.*\]\s+of\s+)(.*)', re.IGNORECASE)


    for var in variables:
        var_name_scl = format_variable_name(var.get("name"))
        var_dtype_raw = var.get("datatype", "VARIANT")
        var_comment = var.get("comment")
        start_value = var.get("start_value")
        children = var.get("children")
        array_elements = var.get("array_elements")

        declaration_dtype = var_dtype_raw
        base_type_for_init = var_dtype_raw
        is_array = False
        is_potential_udt = False
        udt_name = None
        check_type_for_udt = None # El tipo base limpio a comprobar

        if isinstance(var_dtype_raw, str):
            # 1. Es Array?
            array_match = array_pattern.match(var_dtype_raw)
            if array_match:
                is_array = True
                array_prefix_for_decl = array_match.group(1)
                base_type_raw = array_match.group(2).strip()
                base_type_for_init = base_type_raw
                check_type_for_udt = base_type_raw # Comprobar el tipo base

                # Determinar declaración SCL para el array
                if base_type_raw.startswith('"') and base_type_raw.endswith('"'):
                    declaration_dtype = f'{array_prefix_for_decl}{base_type_raw}'
                    check_type_for_udt = base_type_raw[1:-1] # Quitar comillas para comprobar UDT
                elif base_type_raw.lower() not in basic_types and \
                     not string_pattern.match(base_type_raw) and \
                     base_type_raw.lower() != 'char' and base_type_raw.lower() != 'wchar':
                     # Asumir UDT si no es básico conocido y no es String[N]/Char
                     declaration_dtype = f'{array_prefix_for_decl}"{base_type_raw}"' # Poner comillas
                else: # Es básico o String[N]/Char
                     declaration_dtype = f'{array_prefix_for_decl}{base_type_raw}'

            else: # 2. No es Array
                base_type_for_init = var_dtype_raw
                check_type_for_udt = var_dtype_raw # Comprobar el tipo completo

                if var_dtype_raw.startswith('"') and var_dtype_raw.endswith('"'):
                    declaration_dtype = var_dtype_raw
                    check_type_for_udt = var_dtype_raw[1:-1] # Quitar comillas para comprobar UDT
                elif var_dtype_raw.lower() not in basic_types and \
                     not string_pattern.match(var_dtype_raw) and \
                     var_dtype_raw.lower() != 'char' and var_dtype_raw.lower() != 'wchar':
                     # Asumir UDT si no es básico conocido y no es String[N]/Char
                     declaration_dtype = f'"{var_dtype_raw}"' # Poner comillas
                else: # Es básico o String[N]/Char
                     declaration_dtype = var_dtype_raw
                     check_type_for_udt = None # No necesita comprobación de UDT


            # 3. Comprobación final si es potencial UDT (basado en check_type_for_udt)
            if check_type_for_udt:
                if check_type_for_udt.lower() not in basic_types and \
                   not string_pattern.match(check_type_for_udt) and \
                   check_type_for_udt.lower() != 'char' and check_type_for_udt.lower() != 'wchar':
                   # Solo si después de limpiar (quitar comillas, etc.) sigue sin ser básico/paramétrico
                   is_potential_udt = True
                   udt_name = check_type_for_udt # Nombre limpio del UDT

        # --- Buscar archivo de definición SOLO si es_potential_udt ---
        if is_potential_udt and udt_name and project_root_dir:
            udt_scl_name = format_variable_name(udt_name)
            expected_udt_path = os.path.join(project_root_dir, 'PLC data types', 'parsing', f'{udt_scl_name}_processed.json')
            relative_udt_path = os.path.relpath(expected_udt_path, project_root_dir)
            if os.path.exists(expected_udt_path):
                 print(f"    INFO: Definición UDT '{udt_name}' localizada en: '{relative_udt_path}'")
            else:
                 print(f"    WARNING: No se encontró definición para UDT '{udt_name}'. Se esperaba en: '{relative_udt_path}'")

        # --- Construir línea de declaración (resto sin cambios) ---
        declaration_line = f"{indent}{var_name_scl} : {declaration_dtype}"
        init_value_scl_part = ""

        if children:
            scl_lines.append(declaration_line); scl_lines.append(f"{indent}STRUCT")
            scl_lines.extend(generate_scl_declarations(children, indent_level + 1, project_root_dir))
            scl_lines.append(f"{indent}END_STRUCT;")
            if var_comment: scl_lines[-1] += f" // {var_comment}"
            scl_lines.append("")
            continue

        init_value_scl = None
        if is_array and array_elements:
            try: indices_numeric = {int(k): v for k, v in array_elements.items()}; sorted_indices_str = [str(k) for k in sorted(indices_numeric.keys())]
            except ValueError: print(f"Advertencia: Índices array no numéricos para '{var_name_scl}', ordenando como strings."); sorted_indices_str = sorted(array_elements.keys())
            init_values = []
            for idx_str in sorted_indices_str:
                val_info = array_elements[idx_str]
                val_raw = val_info.get('value') if isinstance(val_info, dict) else val_info
                formatted_val = format_scl_start_value(val_raw, base_type_for_init)
                init_values.append(formatted_val if formatted_val is not None else 'NULL')
            if init_values: init_value_scl = f"[{', '.join(init_values)}]"
        elif not is_array and not children and start_value is not None:
            init_value_scl = format_scl_start_value(start_value, base_type_for_init)

        if init_value_scl is not None:
            init_value_scl_part = f" := {init_value_scl}"

        declaration_line += f"{init_value_scl_part};"
        if var_comment: declaration_line += f" // {var_comment}"
        scl_lines.append(declaration_line)

    return scl_lines